@@ -653,19 +653,53 @@ _xdg_mime_magic_get_buffer_extents (XdgMimeMagic *mime_magic)
 const char *
 _xdg_mime_magic_lookup_data (XdgMimeMagic *mime_magic,
 			     const void   *data,
-			     size_t        len)
+			     size_t        len,
+                             const char   *mime_types[],
+                             int           n_mime_types)
 {
   XdgMimeMagicMatch *match;
   const char *mime_type;
+  int n;
+  int priority;
+  int had_match;
 
   mime_type = NULL;
+  priority = 0;
+  had_match = 0;
   for (match = mime_magic->match_list; match; match = match->next)
     {
       if (_xdg_mime_magic_match_compare_to_data (match, data, len))
 	{
-	  if ((mime_type == NULL) || (xdg_mime_mime_type_subclass (match->mime_type, mime_type))) {
-	    mime_type = match->mime_type;
-	  }
+	  if (!had_match || match->priority > priority ||
+	      (mime_type != NULL && xdg_mime_mime_type_subclass (match->mime_type, mime_type)))
+	    {
+	      mime_type = match->mime_type;
+	      priority = match->priority;
+	    }
+	  else if (had_match && match->priority == priority)
+	    /* multiple unrelated patterns with the same priority matched,
+	     * so we can't tell what type this is. */
+	    mime_type = NULL;
+
+	  had_match = 1;
+	}
+      else 
+	{
+	  for (n = 0; n < n_mime_types; n++)
+	    {
+	      if (mime_types[n] && 
+		  xdg_mime_mime_type_equal (mime_types[n], match->mime_type))
+		mime_types[n] = NULL;
+	    }
+	}
+    }
+
+  if (mime_type == NULL)
+    {
+      for (n = 0; n < n_mime_types; n++)
+	{
+	  if (mime_types[n])
+	    mime_type = mime_types[n];
 	}
     }
 
