@@ -84,63 +84,6 @@
 #define DIR_SEPARATORS "/"
 #endif
 
-#if !GLIB_CHECK_VERSION (2,7,0)
-/* g_access lifted from GLib */
-#ifndef G_OS_WIN32
-#define g_access(filename, mode) access (filename, mode)
-#else
-static int
-g_access (const gchar *filename,
-	  int          mode)
-{
-#ifdef G_OS_WIN32
-  if (G_WIN32_HAVE_WIDECHAR_API ())
-    {
-      wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval;
-      int save_errno;
-      
-      if (wfilename == NULL)
-	{
-	  errno = EINVAL;
-	  return -1;
-	}
-
-      retval = _waccess (wfilename, mode);
-      save_errno = errno;
-
-      g_free (wfilename);
-
-      errno = save_errno;
-      return retval;
-    }
-  else
-    {    
-      gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval;
-      int save_errno;
-
-      if (cp_filename == NULL)
-	{
-	  errno = EINVAL;
-	  return -1;
-	}
-
-      retval = access (cp_filename, mode);
-      save_errno = errno;
-
-      g_free (cp_filename);
-
-      errno = save_errno;
-      return retval;
-    }
-#else
-  return access (filename, mode);
-#endif
-}
-#endif
-#endif
-
 #ifdef HAVE_FAM
 static FAMConnection *fam_connection = NULL;
 static gint fam_watch_id = 0;
@@ -779,29 +722,47 @@ read_link (const gchar *full_name)
 
 static void
 get_access_info (GnomeVFSFileInfo *file_info,
-              const gchar *full_name)
+		 const gchar *full_name)
 {
      /* FIXME: should check errno after calling access because we don't
       * want to set valid_fields if something bad happened during one
       * of the access calls
       */
-     if (g_access (full_name, R_OK) == 0) {
-             file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
-     }
-
-     if (g_access (full_name, W_OK) == 0) {
-             file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
-     }
-
 #ifdef G_OS_WIN32
-     if (g_file_test (full_name, G_FILE_TEST_IS_EXECUTABLE)) {
-             file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
-     }
-#else 
-     if (g_access (full_name, X_OK) == 0) {
-             file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
-     }
-#endif 
+	if (g_access (full_name, R_OK) == 0) {
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
+	}
+	if (g_access (full_name, W_OK) == 0) {
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
+	}
+	if (g_file_test (full_name, G_FILE_TEST_IS_EXECUTABLE)) {
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+	}
+#else
+	/* Try to minimize the nr of access calls. We rely on read almost
+	 * always being allowed in normal cases to keep down the number of
+	 * calls needed
+	 */
+	if (g_access (full_name, R_OK|W_OK) == 0) {
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_WRITABLE;
+		if (g_access (full_name, X_OK) == 0) {
+			file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+		}
+	} else if (g_access (full_name, R_OK|X_OK) == 0) {
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+	} else {
+		if (g_access (full_name, R_OK) == 0) {
+			file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
+		} else {
+			if (g_access (full_name, W_OK) == 0) {
+				file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
+			}
+			if (g_access (full_name, X_OK) == 0) {
+				file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+			}
+		}
+	}
+#endif
 
      file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_ACCESS;
 }
@@ -1058,6 +1019,10 @@ do_read_directory (GnomeVFSMethod *method,
 #endif
 	full_name = handle->name_buffer;
 
+	if (handle->options & GNOME_VFS_FILE_INFO_NAME_ONLY) {
+		return GNOME_VFS_OK;
+	}
+		
 	if (get_stat_info (file_info, full_name, handle->options, &statbuf) != GNOME_VFS_OK) {
 		/* Return OK - this should not terminate the directory iteration
 		 * and we will know from the valid_fields that we don't have the
@@ -2481,12 +2446,6 @@ static GnomeVFSResult inotify_monitor_add (GnomeVFSMethod *method,
 					   GnomeVFSMonitorType monitor_type)
 {
 	ih_sub_t *sub;
-	char *filename;
-
-	filename = get_path_from_uri (uri);
-	if (filename == NULL) {
-		return GNOME_VFS_ERROR_INVALID_URI;
-	}
 
 	sub = ih_sub_new (uri, monitor_type);
 	if (sub == NULL) {
@@ -2564,8 +2523,6 @@ static GnomeVFSResult inotify_monitor_cancel (GnomeVFSMethod *method,
 	if (sub->cancelled)
 		return GNOME_VFS_OK;
 
-	sub->cancelled = TRUE;
-
 	ih_sub_cancel (sub);
 	ih_sub_free (sub);
 	return GNOME_VFS_OK;
