@@ -33,6 +33,8 @@
 
 #include <sys/time.h>
 #include <arpa/inet.h>
+
+#include <gconf/gconf-client.h>
 #include <libgnomevfs/gnome-vfs-context.h>
 #include <libgnomevfs/gnome-vfs-inet-connection.h>
 #include <libgnomevfs/gnome-vfs-socket-buffer.h>
@@ -43,6 +45,7 @@
 #include <libgnomevfs/gnome-vfs-module-callback-module-api.h>
 #include <libgnomevfs/gnome-vfs-module.h>
 #include <libgnomevfs/gnome-vfs-module-shared.h>
+#include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-parse-ls.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <stdio.h> /* for sscanf */
@@ -142,9 +145,12 @@ typedef struct {
 	GnomeVFSFileInfoOptions file_info_options;
 } FtpDirHandle;
 
-
+/* Gconf key's "use_http_proxy" is for FTP as well, it should be renamed */
 static const char USE_PROXY_KEY[] = "/system/http_proxy/use_http_proxy";
 
+static const char PROXY_FTP_HOST_KEY[] = "/system/proxy/ftp_host";
+static const char PROXY_FTP_PORT_KEY[] = "/system/proxy/ftp_port";
+
 
 static GnomeVFSResult do_open	         (GnomeVFSMethod               *method,
 					  GnomeVFSMethodHandle         **method_handle,
@@ -179,6 +185,9 @@ static GnomeVFSResult get_list_command (FtpConnection   *conn,
 
 static const int   control_port = 21;
 
+/* FTP Proxy */
+static gchar *proxy_host = NULL;
+static int proxy_port = 0;
 
 /* A GHashTable of FtpConnectionPool */
 
@@ -588,22 +597,43 @@ do_path_command (FtpConnection *conn,
 		 GnomeVFSURI *uri,
 		 GnomeVFSCancellation *cancellation) 
 {
-	char *path;
-	char *actual_command;
+	char *unescaped;
+	gchar *basename,
+	      *path,
+	      *actual_command,
+	      *cwd_command;
 	GnomeVFSResult result;
+	int end;
 
-	/* as some point we may need to make this execute a CD and then
-  	 * a command using the basename rather than the full path. I am yet
-	 * to come across such a system.
-	 */
-	path = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-
-	if (path == NULL || path[0] == '\0') {
-		actual_command = g_strconcat (command, " /", NULL);
-	} else {
-		actual_command = g_strconcat (command, " ", path, NULL);
+	/* Execute a CD and then a command using the basename rather
+	 * than the full path. Useful for some (buggy?) systems. */
+	unescaped = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (unescaped == NULL || unescaped[0] == '\0') {
+		g_free (unescaped);
+		unescaped = g_strdup ("/");
 	}
+
+	/* Remove trailing slashes */
+	end = strlen(unescaped) - 1;
+	if (end > 0 && unescaped[end] == '/') {
+		unescaped[end] = 0;
+	}
+	
+	basename = g_path_get_basename (unescaped);
+	path = g_path_get_dirname (unescaped);
+	g_free (unescaped);
+
+	cwd_command = g_strconcat ("CWD ", path, NULL);
 	g_free (path);
+	result = do_basic_command (conn, cwd_command, cancellation);
+	g_free (cwd_command);
+	if (result != GNOME_VFS_OK) {
+		g_free (basename);
+		return result;
+	}
+
+	actual_command = g_strconcat (command, " ", basename, NULL);
+	g_free (basename);
 
 	result = do_basic_command (conn, actual_command, cancellation);
 	g_free (actual_command);
@@ -736,22 +766,42 @@ do_transfer_command (FtpConnection *conn, gchar *command, GnomeVFSContext *conte
 static GnomeVFSResult
 do_path_transfer_command (FtpConnection *conn, gchar *command, GnomeVFSURI *uri, GnomeVFSContext *context) 
 {
-	char *path;
-	char *actual_command;
+	char *unescaped;
+	gchar *basename,
+	      *path,
+	      *cwd_command,
+	      *actual_command;
 	GnomeVFSResult result;
+	int end;
 
-	/* as some point we may need to make this execute a CD and then
-  	 * a command using the basename rather than the full path. I am yet
-	 * to come across such a system.
-	 */
-	path = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
-
-	if (path == NULL || path[0] == '\0') {
-		actual_command = g_strconcat (command, " /", NULL);
-	} else {
-		actual_command = g_strconcat (command, " ", path, NULL);
+	unescaped = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (unescaped == NULL || unescaped[0] == '\0') {
+		g_free (unescaped);
+		unescaped = g_strdup ("/");
 	}
+
+	/* Remove trailing slashes */
+	end = strlen(unescaped) - 1;
+	if (end > 0 && unescaped[end] == '/') {
+		unescaped[end] = 0;
+	}
+
+	basename = g_path_get_basename (unescaped);
+	path = g_path_get_dirname (unescaped);
+	g_free (unescaped);
+
+	cwd_command = g_strconcat ("CWD ", path, NULL);
 	g_free (path);
+	result = do_basic_command (conn, cwd_command,
+				   get_cancellation (context));
+	g_free (cwd_command);
+	if (result != GNOME_VFS_OK) {
+		g_free (basename);
+		return result;
+	}
+
+	actual_command = g_strconcat (command, " ", basename, NULL);
+	g_free (basename);
 
 	result = do_transfer_command (conn, actual_command, context);
 	g_free (actual_command);
@@ -807,6 +857,13 @@ save_authn_info (GnomeVFSURI *uri,
         g_free (in_args.uri);
 }
 
+/*
+ * Returns FALSE if callback was not handled (*aborted
+ *   will be FALSE, user will be "anonymous")
+ *
+ * Returns TRUE if callback invocation succeeded, *aborted
+ *   will be set to TRUE only if the user didn't cancel
+ */
 static gboolean
 query_user_for_authn_info (GnomeVFSURI *uri, 
 			   char **user, char **pass, char **keyring,
@@ -817,7 +874,7 @@ query_user_for_authn_info (GnomeVFSURI *uri,
         GnomeVFSModuleCallbackFullAuthenticationOut out_args;
 	gboolean ret;
 
-	ret = FALSE;
+	ret = *aborted = FALSE;
 	
 	memset (&in_args, 0, sizeof (in_args));
         memset (&out_args, 0, sizeof (out_args));
@@ -845,7 +902,6 @@ query_user_for_authn_info (GnomeVFSURI *uri,
                                                 &out_args, sizeof (out_args));
 
 	if (!ret) {
-		ret = TRUE;
 		/* No callback, try anon login */
 		*user = g_strdup ("anonymous");
 		*pass = g_strdup ("nobody@gnome.org");
@@ -884,7 +940,7 @@ query_user_for_authn_info (GnomeVFSURI *uri,
        	g_free (out_args.password);
        	g_free (out_args.keyring);
 
-	return ret && !out_args.abort_auth;
+	return ret;
 } 
 
 static gboolean
@@ -1071,7 +1127,15 @@ ftp_login (FtpConnection *conn,
 	gchar *tmpstring;
 	GnomeVFSResult result;
 	
-	tmpstring = g_strdup_printf ("USER %s", user);
+	if (!proxy_host) {
+		tmpstring = g_strdup_printf ("USER %s", user);
+	} else {
+		/* Using FTP proxy */
+		tmpstring = g_strdup_printf ("USER %s@%s",
+					     user,
+					     gnome_vfs_uri_get_host_name (conn->uri));
+	}
+
 	result = do_basic_command (conn, tmpstring, cancellation);
 	g_free (tmpstring);
 
@@ -1095,7 +1159,9 @@ try_connection (GnomeVFSURI *uri,
 	gint port = control_port;
 	const char *host;
 	
-	if (gnome_vfs_uri_get_host_port (uri)) {
+	if (proxy_host) {
+		port = proxy_port;
+	} else if (gnome_vfs_uri_get_host_port (uri)) {
                 port = gnome_vfs_uri_get_host_port (uri);
         }
 
@@ -1103,7 +1169,12 @@ try_connection (GnomeVFSURI *uri,
 	if (*saved_ip != NULL) {
 		host = *saved_ip;
 	} else {
-		host = gnome_vfs_uri_get_host_name (uri);
+		if (!proxy_host) {
+			host = gnome_vfs_uri_get_host_name (uri);
+		} else  {
+			/* Use FTP proxy */
+			host = proxy_host;
+		}
 	}
 
 	if (host == NULL) {
@@ -1325,7 +1396,7 @@ ftp_connection_create (FtpConnectionPool *pool,
 								 &aborted, !uri_has_username); 
 				G_LOCK (connection_pools);
 				pool->num_connections--;
-				if (!ret) {
+				if (aborted) {
 					gnome_vfs_uri_unref (conn->uri);
 					g_string_free (conn->response_buffer, TRUE);
 					g_free (conn);
@@ -1333,11 +1404,7 @@ ftp_connection_create (FtpConnectionPool *pool,
 					g_free (pass);
 					g_free (keyring);
 
-					if (aborted) {
-						return GNOME_VFS_ERROR_CANCELLED;
-					} else {
-						return GNOME_VFS_ERROR_LOGIN_FAILED;
-					}
+					return GNOME_VFS_ERROR_CANCELLED;
 				}
 				g_string_free (conn->response_buffer, TRUE);
                         	conn->response_buffer = g_string_new ("");
@@ -1347,7 +1414,9 @@ ftp_connection_create (FtpConnectionPool *pool,
 				if (result == GNOME_VFS_OK) {
 					break;
 				}
-				if (result != GNOME_VFS_ERROR_LOGIN_FAILED) {
+				if (result != GNOME_VFS_ERROR_LOGIN_FAILED ||
+				    !ret /* if callback was not handled, and anonymous
+					    login failed, don't run into endless loop */) {
 					gnome_vfs_uri_unref (conn->uri);
 					g_string_free (conn->response_buffer, TRUE);
 					g_free (conn);
@@ -1556,7 +1625,7 @@ invalidate_dirlist_cache (GnomeVFSURI *uri)
 	G_LOCK (connection_pools);
 	pool = ftp_connection_pool_lookup (uri);
 	g_hash_table_remove (pool->cached_dirlists,
-			     uri->text);
+			     uri->text != NULL ? uri->text : "/");
 	G_UNLOCK (connection_pools);
 }
 
@@ -2370,7 +2439,7 @@ do_open_directory (GnomeVFSMethod *method,
 	G_LOCK (connection_pools);
 	pool = ftp_connection_pool_lookup (uri);
 	cached_dirlist = g_hash_table_lookup (pool->cached_dirlists,
-					      uri->text);
+					      uri->text != NULL ? uri->text : "/");
 	if (cached_dirlist != NULL) {
 		gettimeofday (&tv, NULL);
 
@@ -2451,7 +2520,7 @@ do_open_directory (GnomeVFSMethod *method,
 	cached_dirlist->read_time = tv.tv_sec;
 	
 	g_hash_table_replace (pool->cached_dirlists,
-			      g_strdup (uri->text),
+			      g_strdup (uri->text != NULL ? uri->text : "/"),
 			      cached_dirlist);
 	
 	G_UNLOCK (connection_pools);
@@ -2573,7 +2642,7 @@ do_read_directory (GnomeVFSMethod *method,
 
 				GNOME_VFS_FILE_INFO_SET_SYMLINK (file_info, TRUE);
 				file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
-				file_info->symlink_name = gnome_vfs_unescape_string (uri->text, "/");
+				file_info->symlink_name = gnome_vfs_unescape_string (uri->text != NULL ? uri->text : "/", "/");
 				
 				g_free (file_info->name);
 				file_info->name = real_name;
@@ -2638,6 +2707,9 @@ do_make_directory (GnomeVFSMethod *method,
 		do_path_command_completely (chmod_command, uri, context,
 			GNOME_VFS_ERROR_ACCESS_DENIED);
 		g_free(chmod_command);
+	} else if (result != GNOME_VFS_ERROR_CANCELLED &&
+		   gnome_vfs_uri_exists (uri)) {
+		result = GNOME_VFS_ERROR_FILE_EXISTS;
 	}
 
 	return result;
@@ -2820,12 +2892,37 @@ GnomeVFSMethod *
 vfs_module_init (const char *method_name, 
 		 const char *args)
 {
+	GConfClient *gclient;
+
 	connection_pools = g_hash_table_new (ftp_connection_uri_hash, 
 					     ftp_connection_uri_equal);
+
+	gclient = gconf_client_get_default ();
+	if (gclient) {
+		if (gconf_client_get_bool (gclient, USE_PROXY_KEY, NULL)) {
+			/* Using FTP proxy */
+			proxy_host = gconf_client_get_string (gclient,
+							      PROXY_FTP_HOST_KEY,
+							      NULL);
+			/* Don't use blank hostname */
+			if (proxy_host &&
+			    *proxy_host == 0) {
+				g_free (proxy_host);
+				proxy_host = NULL;
+			}
+			proxy_port = gconf_client_get_int (gclient,
+							   PROXY_FTP_PORT_KEY,
+							   NULL);
+		} else proxy_host = NULL;
+	}
+
 	return &method;
 }
 
 void
 vfs_module_shutdown (GnomeVFSMethod *method)
 {
+	if (proxy_host) {
+		g_free (proxy_host);
+	}
 }
