@@ -24,8 +24,10 @@
 
 	 */
 
-#include "gnome-vfs.h"
+#include <config.h>
 
+#include <libgnomevfs/gnome-vfs-init.h>
+#include <libgnomevfs/gnome-vfs-ops.h>
 #include <stdio.h>
 #include <stdlib.h>
 
